// naumann@stce.rwth-aachen.de

#include "config.h"
#include "dco.hpp"

template<typename T>
VT<T> g(VT<T> p, T x, VT<T> y);

/// residual of implicit Euler equation
template<typename T>
VT<T> r(VT<T> p, T x, VT<T> y, VT<T> y_prev, T delta_x) {
  // conditions
  assert(x>=0); 
  assert(delta_x>0);
  assert(p.size()==np); 
  assert(y.size()==n); 
  assert(y_prev.size()==n);
  // residual
  return y-y_prev-delta_x*g(p,x,y);
}

/// Jacobian of residual of implicit Euler system by tangent AD
template<typename T>
MT<T> dr_dy(VT<T> p, T x, VT<T> y, VT<T> y_prev, T delta_x) {
  using TT=dco::gt1s<T>::type;
  VT<TT> r_t(n), y_t(n);
  int i=0;
  while (i<n) {
    y_t(i)=y(i);
    i=i+1;
  }
  MT<T> drdy(n,n);
  i=0;
  while (i<n) {
    dco::derivative(y_t(i))=1;
    r_t=r<TT>(p,x,y_t,y_prev,delta_x);
    int j=0;
    while (j<n) {
      drdy(j,i)=dco::derivative(r_t(j));
      j=j+1;
    }
    dco::derivative(y_t(i))=0;
    i=i+1;
  }
  return drdy;
}

/// Newton method for computing root of implicit Euler system
template<typename T>
VT<T> newton(VT<T> p, T x, VT<T> y, T delta_x) {
  // conditions
  assert(x>=0); 
  assert(delta_x>0);
  assert(p.size()==np); 
  assert(y.size()==n); 
  // state computed by previous implicit Euler step
  VT<T> y_prev=y;
  // residual of implicit Euler system
  VT<T> r_=r(p,x,y,y_prev,delta_x);
  do {
    // Jacobian of residual by tangent AD
    MT<T> drdy=dr_dy(p,x,y,y_prev,delta_x);
    // Newton step solves system of linear equations
    y=y+drdy.lu().solve(-r_);
    r_=r(p,x,y,y_prev,delta_x);
  } while (r_.norm()>eps); // convergence criterion
  return y;
}

/// implicit Euler simulation of differential model
template<typename T>
VT<T> f(VT<T> p, T x) { 
  // conditions
  assert(x>=0); assert(p.size()==np);
  // integration step size
  T delta_x=x/m;
  // initial position fixed to zero
  x=0;
  // initial state fixed to zero
  VT<T> y=VT<T>::Zero(n);
  int i=0;
  while (i<m) {
    // implicit Euler step
    y=newton(p,x,y,delta_x); // root of implicit Euler equation
    x=x+delta_x;
    i=i+1;
  }
  return y;
}
