// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"

#include "dco.hpp"

template<typename T>
VT<T> f(VT<T> p, T x);

/// least squares error of model predictions relative to observations
template<typename T>
T e(VT<T> p, VT<T> xobs, MT<T> yobs) {
  // conditions
  assert(p.size()==np);
  assert(xobs.size()==nobs);
  assert(yobs.rows()==nobs);
  assert(yobs.cols()==n);
  // least squares error
  T e_=0;
  int i=0;
  while (i<xobs.size()) {
    e_=e_+(f(p,xobs(i))-yobs.row(i).transpose()).squaredNorm();
    i=i+1;
  } 
  return e_;
}

template<typename T>
VT<T> de_dp(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(p.size()==np); 
  assert(xobs.size()==nobs); 
  assert(yobs.rows()==nobs); 
  assert(yobs.cols()==n); 
  using AT=dco::ga1s<T>::type;
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  AT e_a;
  VT<AT> p_a(np);
  int i=0;
  while (i<np) {
    p_a(i)=p(i);
    tape->register_variable(p_a(i));
    i=i+1;
  }
  e_a=e<AT>(p_a,xobs,yobs);
  tape->register_output_variable(e_a);
  dco::derivative(e_a)=1;
  tape->interpret_adjoint();
  VT<T> de_dp(np);
  i=0;
  while (i<np) {
    de_dp(i)=dco::derivative(p_a(i));
    i=i+1;
  }
  return de_dp;
}

template<typename T>
MT<T> dde_dp_dp(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(p.size()==np); 
  assert(xobs.size()==nobs); 
  assert(yobs.rows()==nobs); 
  assert(yobs.cols()==n); 
  using TT=dco::gt1s<T>::type;
  TT e_t;
  VT<TT> p_t(np);
  int i=0;
  while (i<np) {
    p_t(i)=p(i);
    i=i+1;
  }
  VT<TT> de_dp_t(np);
  MT<T> dde_dp_dp(np,np);
  i=0;
  while (i<np) {
    dco::derivative(p_t(i))=1;
    de_dp_t=de_dp<TT>(p_t,xobs,yobs);
    int j=0;
    while (j<np) {
      dde_dp_dp(j,i)=dco::derivative(de_dp_t(j));
      j=j+1;
    }
    dco::derivative(p_t(i))=0;
    i=i+1;
  }
  return dde_dp_dp;
}

template<typename T>
VT<T> de_dobs(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(p.size()==np); 
  assert(xobs.size()==nobs); 
  assert(yobs.rows()==nobs); 
  assert(yobs.cols()==n); 
  using AT=dco::ga1s<T>::type;
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  AT e_a;
  VT<AT> xobs_a(nobs);
  MT<AT> yobs_a(nobs,n);
  int i,j,k; 
  i=0;
  while (i<nobs) {
    xobs_a(i)=xobs(i);
    tape->register_variable(xobs_a(i));
    j=0;
    while (j<n) {
      yobs_a(i,j)=yobs(i,j);
      tape->register_variable(yobs_a(i,j));
      j=j+1;
    } 
    i=i+1;
  }
  e_a=e<AT>(p,xobs_a,yobs_a);
  tape->register_output_variable(e_a);
  dco::derivative(e_a)=1;
  tape->interpret_adjoint();
  VT<T> de_dobs((n+1)*nobs);
  i=0; k=0;
  while (i<nobs) {
    de_dobs(k)=dco::derivative(xobs_a(i));
    k=k+1;
    j=0;
    while (j<n) {
      de_dobs(k)=dco::derivative(yobs_a(i,j));
      k=k+1;
      j=j+1;
    } 
    i=i+1;
  }
  return de_dobs;
}

template<typename T>
MT<T> dde_dp_dobs(VT<T> p, VT<T> xobs, MT<T> yobs) {
  assert(p.size()==np); 
  assert(xobs.size()==nobs); 
  assert(yobs.rows()==nobs); 
  assert(yobs.cols()==n); 
  using TT=dco::gt1s<T>::type;
  TT e_t;
  VT<TT> p_t(np);
  int i,j;
  i=0;
  while (i<np) {
    p_t(i)=p(i);
    i=i+1;
  }
  VT<TT> de_dobs_t((n+1)*nobs);
  MT<T> dde_dobs_dp((n+1)*nobs,np);
  i=0;
  while (i<np) {
    dco::derivative(p_t(i))=1;
    de_dobs_t=de_dobs<TT>(p_t,xobs,yobs);
    j=0;
    while (j<(n+1)*nobs) {
      dde_dobs_dp(j,i)=dco::derivative(de_dobs_t(j));
      j=j+1;
    }
    dco::derivative(p_t(i))=0;
    i=i+1;
  }
  return dde_dobs_dp.transpose();
}
